//Author : Steven Keith can be blamed for this...

#include <string>
#include <fstream>
#include <cassert>
#include <map>
#include <list>
#include <set>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include "hint.h"
#include "bucket.h"

using namespace std;

template<class DataType>
class LinearHashCubeIterator; 
template<class DataType>
class RangeBucketIterator;


template<class DataType>
class LinearHashCube:public MultidimensionalArray<DataType> {
    
 public:
    LinearHashCube( int NumberOfDimensions, string filename = "LinearHashCubeTemp.bin") throw(std::ifstream::failure);

    virtual ~LinearHashCube(){}; 
    
    string getName() const {return this->mFilename;}

    int getNumberOfDimensions() const {return this->mNumberOfDimensions;}
    
    int getNumberOfBuckets() const {return this->storage.size();}

    //this does nothing right now
    void open() throw(std::ifstream::failure){;}
    
    //does nothing
    void close(){;}
    
    //does nothing
    void sync(){;}
    
    //deletes all values in the cube
    void fillWithZeros();

    //inserts a value into the cube
    inline void put(const vector<int>& pos, DataType value);
    inline void put(const int pos[], DataType value);
    
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;
    
    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;
    
    inline int getNumberOfRecords() const {return mNumRecords;}
    
//    void printData();

    BucketIterator<DataType> getIter(int x) const {return *storage[x].iterator();}
    Bucket<DataType> getBucket(int i) const{return storage[i];}

    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new LinearHashCubeIterator<DataType>(*this,start,end));
    }

    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator() const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new LinearHashCubeIterator<DataType>(*this));
    }


    //the following functions should likely be private... will move them
    //when testing is complete.

    int interleave(const vector<int> key) const;
    int interleave(const int key[]) const;
	
    inline float measureUsage(){
	return (float)this->mNumRecords/(float)(this->storage.size()*this->mRecordsPerBucket);
    }

 protected:
    
 private:

    void split();

    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape,string file, Hint hint = Hint()) const  throw(std::ifstream::failure){
	return new LinearHashCube(shape.size(),file);
	}

    int mNumberOfDimensions;
    Hint mHint;
    string mFilename;
    
    string PREFIX;
    enum {verbose=1};

    vector<Bucket<int> > storage;
    int mNumRecords, mRecordsPerBucket;
    float mTau;
    int s;

    friend class LinearHashCubeIterator<DataType>; 
    friend class RangeBucketIterator<DataType>;

};


template<class DataType>
LinearHashCube<DataType>::LinearHashCube( int NumberOfDimensions, string filename) throw(std::ifstream::failure):MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)), storage(vector<Bucket<int> >(0)){
    
    this->PREFIX = "[lhwrbi.h] ";
    this->mFilename = filename;
    this->mNumberOfDimensions = NumberOfDimensions;
    this->mNumRecords = 0;
    this->mRecordsPerBucket = 100;
    this->mTau = 0.75;
    this->storage.push_back(Bucket<int>(this->mNumberOfDimensions, this->mRecordsPerBucket, 0,0,this->mFilename));
    this->s = 0;
}



template<class DataType>
void LinearHashCube<DataType>::fillWithZeros(){
    this->storage.clear();
    this->storage.push_back(Bucket<int>(this->mNumberOfDimensions, this->mRecordsPerBucket, 0,0,this->mFilename));
    this->mNumRecords = 0;
    this->s = 0;
}

template<class DataType>
int LinearHashCube<DataType>::interleave(const vector<int> key) const{
    vector<int> key2(mNumberOfDimensions);
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key[i];
    }
	
    int val = 0;
    int precision = (int) ceil(30.0 / (float)this->mNumberOfDimensions);
    int max = (2 << precision >> 1);
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key2[i] % max;
    }
    int count = 29;

    for(int i = 0; i < precision; i++){
	for(int j = this->mNumberOfDimensions - 1; j >= 0; j--){
	    if((precision - i) * this->mNumberOfDimensions - (this->mNumberOfDimensions - j) > count) continue;
	    val += ((key2[j] % 2)* (2 << count >> 1));
	    key2[j] = key2[j] / 2;
	    count--;
	}
    }
    return val;
}

template<class DataType>
int LinearHashCube<DataType>::interleave(const int key[]) const{
    vector<int> key2(mNumberOfDimensions);
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key[i];
    }
	
    int val = 0;
    int precision = (int) ceil(30.0 / (float)this->mNumberOfDimensions);
    int max = 2 << precision >> 1;
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key2[i] % max;
    }
    int count = 29;

    for(int i = 0; i < precision; i++){
	for(int j = this->mNumberOfDimensions - 1; j >= 0; j--){
	    if((precision - i) * this->mNumberOfDimensions - (this->mNumberOfDimensions - j) > count) continue;
	    val += ((key2[j] % 2)*(2 << count >> 1));
	    key2[j] = key2[j] / 2;
	    count--;
	}
    }
    return val;
}

template<class DataType>
void LinearHashCube<DataType>::put(const vector<int>& pos, DataType value){
    int val = interleave(pos);
    int index = val % (2 << storage[0].getHash() >> 1);
    int numBuckets = this->storage.size();
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
//    cout << "Index = " << index << endl;
    if(storage[index].insert(pos,value)){
//	cout << "Inserting into bucket " << index << endl;
	this->mNumRecords++;
    }
    if(measureUsage() > this->mTau){
//	cout << "SPLIT!" << endl;
	split();
    }
}

template<class DataType>
void LinearHashCube<DataType>::put(const int pos[], DataType value){
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->put(key,value);
}
   
template<class DataType>
DataType LinearHashCube<DataType>::get(const vector<int>& pos) const{
    int val = interleave(pos);
//    cout << "got here " << storage[0].getHash() << endl;
    int index = val % (2 << (storage[0].getHash() >> 1));
//    cout << "index = " << index << endl;
    int numBuckets = this->storage.size();
//    cout << index << " " << numBuckets << endl;
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
    return storage[index].get(pos);
}

template<class DataType>
DataType LinearHashCube<DataType>::get(const int pos[]) const{
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->get(key);
}

template<class DataType>
bool LinearHashCube<DataType>::undefinedAt(const vector<int>& pos) const{
    int val = interleave(pos);
    int index = val % (2 << storage[0].getHash() >> 1);
    int numBuckets = this->storage.size();
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
    return storage[index].undefinedAt(pos);
}

template<class DataType>
bool LinearHashCube<DataType>::undefinedAt(const int pos[]) const{
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->undefinedAt(key);
}

template<class DataType>
void LinearHashCube<DataType>::split(){
    int hash = this->storage[this->s].getHash() + 1;

    if(hash > 29){
	cerr << "Point's exceed hashing capabilities" << endl;
	cerr << "Consider using a larger bucket size!" << endl;
	return;
    }
    
    Bucket<DataType> splitMe(this->storage[this->s]);
//    cout << "I have to assign " << splitMe.getNumItems() << endl;

    Bucket<DataType> one(Bucket<DataType>(this->mNumberOfDimensions, this->mRecordsPerBucket, hash, this->s, this->mFilename));    
    Bucket<DataType> two(Bucket<DataType>(this->mNumberOfDimensions, this->mRecordsPerBucket, hash, this->storage.size(), this->mFilename));
    
//    storage[s] = Bucket<DataType>(this->mNumberOfDimensions, this->mRecordsPerBucket, hash);
//    storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->mRecordsPerBucket, hash));
    
//    int index1 = s;
//    int index2 = this->storage.size() - 1;
    
    std::auto_ptr<BucketIterator<DataType> > it(splitMe.iterator());
    
    //splitting the contents between the two buckets
    while(it->next()){
	
	if(interleave(it->key()) % (2 << hash >> 1) == s){
//	    this->storage[index1].insert(it->key(), it->val());
	    
	    one.insert(it->key(), it->val());
//	    cout << "one" << endl;
	}
	else{
//	    this->storage[index2].insert(it->key(), it->val());
	    two.insert(it->key(), it->val());
//	    cout << "two" << endl;
	}
    }//end while

    storage[s] = one;
    storage.push_back(two);

    //Check to make sure that good things happened.
//    cout << "One = " << storage[s].getNumItems() << " Two = " << storage[storage.size() -1].getNumItems() << endl;
    
    //incremement s or reset it to 0
    s++;
    if(s == (2 << hash >> 2)){
	s = 0;
    }
    return;
}


template <class DataType>
class LinearHashCubeIterator:public MultidimensionalIterator<DataType> {

 public:
    LinearHashCubeIterator(const LinearHashCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
	MultidimensionalIterator<DataType>(),
	mStart(start), 
	mEnd(end),
	mOwner(owner),  
	mKey(mStart), 
	mCurrentValue(0),
	iter(0),
	mSearchTheseBuckets(0)
	{
	    RangeBucketIterator<DataType> rbi(this->mOwner, this->mStart, this->mEnd);
	    list<int> temp = rbi.getBucketList();
	    mSearchTheseBuckets.splice(mSearchTheseBuckets.end(),temp);
	    mSearchTheseBuckets.push_back(-1);
	    list<int>::const_iterator it(mSearchTheseBuckets.begin());
/*	    cout<< "Searching these buckets" << endl;
	    while(it != mSearchTheseBuckets.end()){
		cout << "bucket #" << *it << endl;
		it ++;
		}*/
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    if(currentBucket >= 0){
		this->iter= mOwner.getIter(currentBucket);
	    }
	}

    LinearHashCubeIterator(const LinearHashCube<DataType> & other):
	MultidimensionalIterator<DataType>(),
	mStart(other.getShape().size()),
	mEnd(other.getShape().size()), 
	mOwner(other), 
	mKey(other.getShape().size()),
	mCurrentValue(0),
	iter(0),
	mSearchTheseBuckets(0)
	{
	    for(uint i = 0; i < mStart.size(); i++){
		this->mStart[i] = 0;
	    }
	    for(uint i = 0; i < mEnd.size(); i++){
		this->mEnd[i] = 99999999;
	    }
	    RangeBucketIterator<DataType> rbi(this->mOwner, this->mStart, this->mEnd);
	    for(int i = 0; i < this->mOwner.getNumberOfBuckets(); i++){
		this->mSearchTheseBuckets.push_back(i);
	    }
	    this->mSearchTheseBuckets.push_back(-1);
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    if(currentBucket >= 0){
		this->iter= mOwner.getIter(currentBucket);
	    }

	}
    
    LinearHashCubeIterator(const LinearHashCubeIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
	mStart(other.mStart), 
	mEnd(other.mEnd), 
	mOwner(other.mOwner), 
	mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),
	iter(other.iter),
	mSearchTheseBuckets(0)
	{
	    this->currentBucket = 0;
	    list<int> temp = other.mSearchTheseBuckets;
	    mSearchTheseBuckets.splice(mSearchTheseBuckets.end(),temp);
	}
    
    
    inline bool next(){
	if(mSearchTheseBuckets.size() == 0)
	    return false;
	
	bool alright = false;
	if(iter.next()){
	    while(!alright){
		alright = true;
		for(int i = 0; i < mOwner.getNumberOfDimensions(); i++){
		    int temp = iter.key(i);
		    if(!(mStart[i] <= temp && temp <= mEnd[i])){
			alright = false;
		    }
		}
		if(alright){
		    mCurrentValue = iter.val();
		    for(int i = 0; i < mOwner.getNumberOfDimensions(); i++){
			mKey[i] = iter.key(i);
		    }
		}
		else if(!iter.next()){
		    return this->next();
		}
	    }
	}
	else{
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    while( this->mSearchTheseBuckets.size() > 0 && this->mOwner.storage[currentBucket].getNumItems() == 0){
		this->currentBucket = *mSearchTheseBuckets.begin();
		this->mSearchTheseBuckets.pop_front();
	    }
	    if(mSearchTheseBuckets.size() == 0) return false;
	    
	    iter= mOwner.getIter(currentBucket);
	    mCurrentValue = iter.val();
	    for(int i = 0; i < mOwner.getNumberOfDimensions(); i++){
		mKey[i] = iter.key(i);
	    }	
	}
	return true;
    }
    
    inline DataType value() const{
	return this->mCurrentValue;
    }
    
    inline vector<int>& key(){
	return mKey;
    }
    
    LinearHashCubeIterator<DataType> * clone() const {return new LinearHashCubeIterator<DataType>(*this);}
    
 protected:
    int currentBucket;
    vector<int> mStart, mEnd;
    const LinearHashCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    enum {verbose=0};
    BucketIterator<DataType> iter;
    list<int> mSearchTheseBuckets;
    
    
};

template<class DataType>
class RangeBucketIterator{

 public:
    
    RangeBucketIterator(const LinearHashCube<DataType> & owner,const vector<int>& start, const vector<int>& end): 
	mStart(start), 
	mEnd(end),
	mBuckets(0),
	mOwner(owner)
	{
	    this->query(0,1,mStart,mEnd);
	}
    
    
    void query(int index, int hash, vector<int> ll, vector<int> ur);

    list<int> getBucketList() const{return mBuckets;}
    
 protected:
    vector<int> mStart, mEnd;
    list<int> mBuckets;
    const LinearHashCube<DataType> & mOwner;

};

template<class DataType>
void RangeBucketIterator<DataType>::query(int index, int hash, vector<int> in_ll, vector<int> in_ur){
    list<int> returnMe(0);
    if(this->mOwner.getBucket(index).getHash() < hash){
	returnMe.push_back(index);
	this->mBuckets.splice(mBuckets.end(),returnMe);
	return;
    }
    int ll = this->mOwner.interleave(in_ll);
    int ur = this->mOwner.interleave(in_ur);
    
    int h1 = (ll % (2 << hash >>1) - ll % (2 << hash >> 2))/(2 << hash >> 2);
    int h2 = (ur % (2 << hash >>1) - ur % (2 << hash >> 2))/(2 << hash >> 2);
    
    //Does the lower left and upper right corners of the query rectangle
    //hash to the same bucket?
    if(h1 == h2){
	this->query(index + h1 * (2 << hash >> 2),hash+1,in_ll,in_ur);
	return;

    }//otherwise redefine each of the queries to fit in a single bucket
    else{
	int numDims = this->mOwner.getNumberOfDimensions();
	vector<int> ur2(numDims), ll2(numDims);
	for(int i = 0; i < numDims; i++){
	    ur2[i] = in_ur[i];
	    ll2[i] = in_ll[i];
	}

	int coordinate = (hash-1) % numDims;
	int blockSize = (2 << (30/numDims - (hash-1)/numDims) >> 1);
	ur2[coordinate] += (ur2[coordinate] + blockSize) % blockSize;
	ll2[coordinate] -= ll2[coordinate] % blockSize;
	this->query(index,hash+1,in_ll,ur2);
	this->query(index+(int)(2 << hash >> 2),hash+1,ll2,in_ur);
	return;
    }
}
