#include "lhwrbi.h"
#include "hint.h"
#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>

void hardertest(){
    srand(0);
    int numDims = 5;
    int numBits = 30;
    int numInserts = 1000000;
    vector<int> pos(numDims);
    
    LinearHashCube<int> cube(numDims, "linearhashtest.bin");
    
    while(cube.getNumberOfRecords() < numInserts){
	for(int j = 0; j < numDims; j++){
	    pos[j] =(int)floor((double)rand()/RAND_MAX*(2<<(numBits/numDims)>>1));
//	    cout << pos[j] << endl;
	}
	cube.put(pos,1);
    }
}

void simpletest(){
    int numDims = 2;
    
    cout << "Creating Linear Hash Cube... ";
    LinearHashCube<int> cube(numDims, "linearhashtest.bin");
    cout << "done."<< endl;
    
    cout << "Getting the name of the cube... " << cube.getName() << endl;
    
    cout << "Filling the cube with zeros... ";
    cube.fillWithZeros();
    cout << "done."<< endl;
    
    vector<int> key(2);
    key[0] = 101;
    key[1] = 5;
    cout <<"Testing the interleave function on "<<key[0]<<", "<<key[1]<<"...";
    int val = cube.interleave(key);
    cout << val << endl;;
    
    cout << "Trying a put... ";
    cube.put(key,1);
    cout << "done."<<endl;

    cout << "Testing the getNumberOfRecords function... ";
    int numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    cout << "Trying a get... ";
    int x = cube.get(key);
    cout << "I got "<< x << endl;
    
    
    cout << "Trying a put over the previous put... ";
    cube.put(key,2);
    cout << "done."<<endl;
    
    cout << "Testing the getNumberOfRecords function (should still be 1)... ";
    numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    cout << "Trying a get... ";
    x = cube.get(key);
    cout << "I got "<< x << endl;
    
    key[1] = 6;
    cout << "Trying another put in a different location... ";
    cube.put(key,3);
    cout << "done."<<endl;
    
    cout << "Trying a get... ";
    x = cube.get(key);
    cout << "I got "<< x << endl;
    
    cout << "Testing the getNumberOfRecords function (should be 2)... ";
    numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    key[0] = 0;
    cout << "Trying another put in a different location... ";
    cube.put(key,5);
    cout << "done."<<endl;
    
    cout << "Trying undefinedAt " << key[0] << ", " << key[1]<<"... ";
    bool y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    key[0] = 10;
    cout << "Trying undefinedAt " << key[0] << ", " <<key[1]<< "... ";
    y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    key[0] = 101;
    key[1] = 5;
    cout << "Trying undefinedAt " << key[0] << ", " <<key[1]<< "... ";
    y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    vector<int> start(2);
    start[0] = 0;
    start[1] = 5;
    vector<int> end(2);
    end[0] = 250;
    end[1] = 5;
    MultidimensionalArray<int>::IteratorPtr iter(cube.iterator(start,end));
    while(iter->next()){
	cout << iter->value() << endl;
    }
    
}

int main() {
    simpletest();
}

