//Author : Steven Keith can be blamed for this...


#include <pqxx/pqxx>
#include <pqxx/connection>
#include <pqxx/tablewriter>
#include <pqxx/transaction>
#include <pqxx/result>
#include <string>
#include <fstream>
#include <cassert>
#include <map>
#include <set>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <map>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include "hint.h"
#include "bulkwriter.h"

using namespace std;
using namespace PGSTD;
using namespace pqxx;

template<class DataType>
class PostgreSQLCubeIterator; 
template<class DataType>
class PostgreSQLCubeSliceIterator;
template<class DataType>
class PostgreSQLCubeDiceIterator;


template<class DataType>
class PostgreSQLCube:public MultidimensionalArray<DataType> {
    
 public:
    PostgreSQLCube( int NumberOfDimensions, string filename = "BTreeCubeTemp.bin") throw(std::ifstream::failure);
    //should finish loading anything in the queue
    //deleting everything right now while developing 
    virtual ~PostgreSQLCube(){}; 
    
    string getName() const;
    
    void open() throw(std::ifstream::failure);
    
    void close(); 
    
    void sync();
    
    void fillWithZeros();

    inline void put(const vector<int>& pos, DataType value);
    inline void put(const int pos[], DataType value);
    
    inline void setLoading(bool b){
	loading = b;
	if(!b){
	    bw.commit();
	}
    }
    
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;
    
    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;
    
    int getNumberOfRecords();
    
    void printData();
    
    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
	return typename MultidimensionalArray<DataType>::IteratorPtr(new PostgreSQLCubeIterator<DataType>(*this,start,end));
    }
    
    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator() const {
	return typename MultidimensionalArray<DataType>::IteratorPtr(new PostgreSQLCubeIterator<DataType>(*this));
    }
    
    typename MultidimensionalArray<DataType>::IteratorPtr sliceIterator( int dim, int slice_at) {
	return new PostgreSQLCubeSliceIterator<DataType>(*this,dim,slice_at);
    }
    
    typename MultidimensionalArray<DataType>::IteratorPtr diceIterator(  int dim, int stop_after) {
	return new PostgreSQLCubeDiceIterator<DataType>(*this, dim, stop_after);
    }
    
    
    
 protected:
    
 private:
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape,string file, Hint hint = Hint()) const  throw(std::ifstream::failure); 
    
    int mNumberOfDimensions;
    Hint mHint;
    string mFilename;
    
    string PREFIX;
    enum {verbose=0};
    
    bool active;
    bool loading;
    lazyconnection C;
    BulkWriter<int> bw;
    
    
    friend class PostgreSQLCubeIterator<DataType>; 
    friend class PostgreSQLCubeSliceIterator<DataType>;
  friend class PostgreSQLCubeDiceIterator<DataType>;

};


template<class DataType>
PostgreSQLCube<DataType>::PostgreSQLCube( int NumberOfDimensions, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)),C("dbname=lemur user=steve"), bw(filename){
    
    this->PREFIX = "[psqlcube.h] ";
    this->mFilename = filename;
    this->mNumberOfDimensions = NumberOfDimensions;
    this->active = false;
    this->loading = false;
    this->open();
}

template<class DataType>
void PostgreSQLCube<DataType>::open() throw(std::ifstream::failure){

    this->active = true;

    try{
	work T(this->C, "psqlcube.h -- open()");
	if(verbose) cout<<this->PREFIX<<"checking to see if table "<<this->mFilename<<" exists..."<<endl;
	T.exec(("SELECT COUNT(measure) FROM " + this->mFilename).c_str() );
	//shouldn't get here unless the table filename exists... 
	//if it exists do nothing.
	this->bw.getBulkWriter();

	return;

	//throw std::ifstream::failure("This table already exists");
    }catch(const sql_error &e){
	if(verbose) cout<< this->PREFIX<<"Good! Table " << this->mFilename << " does not exist"<<endl;
    }

    if(verbose) cout<<this->PREFIX<<"Creating the table " << this->mFilename <<" with " << this->mNumberOfDimensions<<" dimensions..." << endl;
    try{
	stringstream dims;
	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims<< "d" << i << " INTEGER, ";
	    dims2<< "d" << i << "=new.d"<< i;
	    if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TABLE " + this->mFilename + "(" + dims.str() + "measure INTEGER);").c_str() );
	    T.commit();
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE FUNCTION " + this->mFilename + "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + this->mFilename + " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	    T.commit();
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TRIGGER " + this->mFilename + "_trigger BEFORE INSERT ON " + this->mFilename + " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename + "_function();").c_str());
	    T.commit();
	}
	this->bw.getBulkWriter();

//	C.disconnect();
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	throw std::ifstream::failure("SQL Error creating table");
    }
}



template<class DataType>
MultidimensionalArray<DataType>* PostgreSQLCube<DataType>::createNew(vector<int> shape, string file, Hint h) const  throw(std::ifstream::failure){
   return new PostgreSQLCube<DataType>(shape.size(),file);
}



template<class DataType>
string PostgreSQLCube<DataType>::getName() const{
    return this->mFilename;
}

template<class DataType>
void PostgreSQLCube<DataType>::sync(){
    return;
}

template<class DataType>
void PostgreSQLCube<DataType>::close(){
    this->active = false;
    this->bw.deleteBulkWriter();
    // Set up a connection to the backend.
//    connection C("dbname=lemur user=steve");

    if(verbose) cout<<this->PREFIX<<"Dropping the table " << this->mFilename << endl;
    try{
	
	work T(this->C, "psqlcube.h -- close()");
    	T.exec(("DROP TABLE " + this->mFilename + ";").c_str() );
    	T.exec(("DROP FUNCTION " + this->mFilename + "_function() CASCADE;").c_str() );
	T.commit();
//	C.disconnect();
    }catch(const sql_error &e){
	cout<<"getting here"<<endl;
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	throw std::ifstream::failure("SQL Error creating table");
    }
    
}

template<class DataType>
void PostgreSQLCube<DataType>::fillWithZeros(){
//    connection C("dbname=lemur user=steve");

    if(verbose) cout<<this->PREFIX<<"Deleting all values from " << this->mFilename << endl;
    try{
	work T(this->C, "psqlcube.h -- fillWithZeros()");
    	T.exec(("DELETE FROM " + this->mFilename + ";").c_str() );
	T.commit();
//	C.disconnect();
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
    }
}

template<class DataType>
void PostgreSQLCube<DataType>::put(const vector<int>& pos, DataType value){
    if(loading){
	bw.insert(pos,value,this->mNumberOfDimensions);
    }else{
	stringstream dims;
	dims << "INSERT INTO " << this->mFilename <<"(";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << "d" << i << ", ";
	}
	dims << "measure) VALUES (";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << pos[i] << ", ";
	}
	dims << value << ");";
//	connection C("dbname=lemur user=steve");
	
	if(verbose) cout<<this->PREFIX<<"INSERTING the following statement into table " << this->mFilename << endl;
	if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
	try{
	    work T(this->C, "psqlcube.h -- put()");
	    T.exec((dims.str()).c_str() );
	    T.commit();
//	    C.disconnect();
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }

}

template<class DataType>
void PostgreSQLCube<DataType>::put(const int pos[], DataType value){
    if(loading){
	bw.insert(pos,value,this->mNumberOfDimensions);
    }else{
	stringstream dims;
	dims << "INSERT INTO " << this->mFilename <<"(";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << "d" << i << ", ";
	}
	dims << "measure) VALUES (";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << pos[i] << ", ";
	}
	dims << value << ");";
//	connection C("dbname=lemur user=steve");
	
	if(verbose) cout<<this->PREFIX<<"INSERTING the following statement into table " << this->mFilename << endl;
	if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
	try{
	    work T(this->C, "psqlcube.h -- put()");
	    T.exec((dims.str()).c_str() );
	    T.commit();
//	    C.disconnect();
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
}

template<class DataType>
DataType PostgreSQLCube<DataType>::get(const vector<int>& pos) const{
//    if(this->bw.getNum() > 0) this->bw.commit();

    stringstream dims;
    dims << "SELECT measure FROM " << this->mFilename;
    dims << " WHERE ";
    for(int i = 0; i < this->mNumberOfDimensions - 1; i++){
	dims << "d" << i << "="<<pos[i] << " AND ";
    }
    dims << "d" <<(this->mNumberOfDimensions-1)<< "=" << pos[this->mNumberOfDimensions-1]<<";";
    connection C("dbname=lemur user=steve");
    if(verbose) cout<<this->PREFIX<<"Getting the value at the following position in the table " << this->mFilename << endl;
    if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
    try{
	work T(C, "psqlcube.h -- get()");
	result R = T.exec((dims.str()).c_str() );
	T.commit();
	C.disconnect();
	if(R.size() == 0) return 0;
	stringstream sval;
	sval<< R[0][0];
	DataType val;
	sval >> val;
	return val;

    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }
    
}

template<class DataType>
DataType PostgreSQLCube<DataType>::get(const int pos[]) const{
//    if(this->bw.getNum() > 0) this->bw.commit();

    stringstream dims;
    dims << "SELECT measure FROM " << this->mFilename;
    dims << " WHERE ";
    for(int i = 0; i < this->mNumberOfDimensions - 1; i++){
	dims << "d" << i << "="<<pos[i] << " AND ";
    }
    dims << "d" <<(this->mNumberOfDimensions-1)<< "=" << pos[this->mNumberOfDimensions-1]<<";";
    connection C("dbname=lemur user=steve");
    if(verbose) cout<<this->PREFIX<<"Getting the value at the following position in the table " << this->mFilename << endl;
    if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
    try{
	work T(C, "psqlcube.h -- get()");
	result R = T.exec((dims.str()).c_str() );
	T.commit();
	C.disconnect();
	if(R.size() == 0) return 0;
	stringstream sval;
	sval<< R[0][0];
	DataType val;
	sval >> val;
	return val;

    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }
    
}


//The undefinedAt functions are INCORRECT... they do not actually work the way
//I know Owen wants them to...
template<class DataType>
bool PostgreSQLCube<DataType>::undefinedAt( const vector<int>& pos) const{
    
    if(this->get(pos) == 0)
	return true;
    else
	return false;
    
}

template<class DataType>
bool PostgreSQLCube<DataType>::undefinedAt( const int pos[]) const{
    if(this->get(pos) == 0)
	return true;
    else
	return false;

}

template<class DataType>
int PostgreSQLCube<DataType>::getNumberOfRecords(){
//    connection C("dbname=lemur user=steve");
    
    if(this->bw.getNum() > 0) this->bw.commit();

    try{
	work T(this->C, "psqlcube.h -- getNUmberOfRecords()");
	if(verbose) cout<<this->PREFIX<<"Getting the number of records in table  "<<this->mFilename<<" exists..."<<endl;
	result R = T.exec(("SELECT COUNT(measure) FROM " + this->mFilename).c_str() );
	T.commit();
//	C.disconnect();
	
	if(R.size() == 0) return 0;
	stringstream sval;
	sval<< R[0][0];
	int val;
	sval >> val;
	return val;
	
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }
    
}

template<class DataType>
void PostgreSQLCube<DataType>::printData(){
    std::auto_ptr<PostgreSQLCubeIterator<DataType> > it(new PostgreSQLCubeIterator<DataType>(*this));
    
    while(it->next()){
	vector<int> key = it->key();
	for(uint i = 0; i < key.size(); i++){
	    cout<< key[i] << " ";
	}
	cout << it->value() << endl;
    }
    
}



template <class DataType>
class PostgreSQLCubeIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeIterator(const PostgreSQLCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
      MultidimensionalIterator<DataType>(),
	    index(-1),R(),
	    mStart(start), mEnd(end),mOwner(owner),  mKey(mStart), mCurrentValue(0), PREFIX ("[PostgreSQLCubeIterator] ")
{
	this->getValues();
	    }

    PostgreSQLCubeIterator(const PostgreSQLCube<DataType> & other):
      MultidimensionalIterator<DataType>(),
	    index(-1),R(),
	    mStart(other.getShape().size()),mEnd(other.getShape().size()), mOwner(other), mKey(other.getShape().size()),mCurrentValue(0),
	PREFIX ("[PostgreSQLCubeIterator] ")
{
	for(uint i = 0; i < mStart.size(); i++){
	    this->mStart[i] = 0;
	}
	for(uint i = 0; i < mEnd.size(); i++){
	    this->mEnd[i] = 99999999;
	}
	this->getValues();

    }

    PostgreSQLCubeIterator(const PostgreSQLCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      index(other.index),R(other.R),
      mStart(other.mStart), mEnd(other.mEnd), mOwner(other.mOwner), mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),  
	PREFIX( "[PostgreSQLCubeIterator] ")
{
}


    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeIterator<DataType> * clone() const {return new PostgreSQLCubeIterator<DataType>(*this);}
 protected:
    int index;
    result R;
    vector<int> mStart, mEnd;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=steve");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    for(uint i = 0; i < this->mStart.size(); i++){
		dims << "d" << i << ">=" << this->mStart[i] << " AND ";
	    } 
	    for(uint i = 0; i < this->mEnd.size()-1; i++){
		dims << "d" << i << "<=" << this->mEnd[i] << " AND ";
	    }

	    dims << "d" << (this->mEnd.size() - 1) << "<=" << this->mEnd[mEnd.size()-1] << ";";
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};

template <class DataType>
class PostgreSQLCubeSliceIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeSliceIterator(const PostgreSQLCube<DataType> & owner,int dim, int at) :
      MultidimensionalIterator<DataType>(), 
	    index(-1),
	    mDim(dim),
	    mAt(at),R(),
	    mOwner(owner), mKey(owner.getShape().size()), mCurrentValue(0),
	PREFIX ( "[PostgreSQLCubeSliceIterator] ")
	    
    {
	this->getValues();
    }

    
    PostgreSQLCubeSliceIterator(const PostgreSQLCubeSliceIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
        index(other.index),
	mDim(other.mDim),
	    mAt(other.mAt), R(other.R),
	    mOwner(other.mOwner), mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),
	    PREFIX ( "[PostgreSQLCubeSliceIterator] ") {
    }
    
    
    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeSliceIterator<DataType> * clone() const {return new PostgreSQLCubeSliceIterator<DataType>(*this);}
 protected:

    int index, mDim, mAt;
    result R;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=steve");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    dims << "d" << this->mDim << "=" << this->mAt << ";";
	    if(verbose) cout << this->PREFIX << dims.str()<<endl;
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};

template <class DataType>
class PostgreSQLCubeDiceIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeDiceIterator(const PostgreSQLCube<DataType> & owner,int dim, int stop_at) :
      MultidimensionalIterator<DataType>(), index(-1), 
	    mDim(dim),mStop_At(stop_at), R(), 
	    mOwner(owner), mKey(owner.getShape().size()), 
	    mCurrentValue(0),PREFIX ("[PostgreSQLCubeDiceIterator] "){
	this->getValues();
    }

    
    PostgreSQLCubeDiceIterator(
		    const PostgreSQLCubeDiceIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
	index(other.index),mDim(other.mDim),mStop_At(other.mStop_At),
	R(other.R),mOwner(other.mOwner),  mKey(other.mKey),
	mCurrentValue(other.mCurrentValue),
    	PREFIX ("[PostgreSQLCubeDiceIterator] ") {
    }
    
    
    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeDiceIterator<DataType> * clone() const {return new PostgreSQLCubeDiceIterator<DataType>(*this);}
 protected:

    int index, mDim, mStop_At;
    result R;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=steve");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    dims << "d" << this->mDim << "<=" << this->mStop_At << ";";
	    if(verbose) cout << this->PREFIX << dims.str()<<endl;
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};
